from datetime import datetime, date

from royal_mail_click_and_drop import (
    AddressRequest as AddressRM,
    BillingDetailsRequest,
    PostageDetailsRequest,
    ShipmentPackageRequest,
)
from royal_mail_click_and_drop.models.address import RecipientDetailsRequest as ContactRM
from royal_mail_click_and_drop.models.shipment_package_request import PackageFormat
from royal_mail_click_and_drop.v2.consts import SendNotifcationsTo

from shipaw.models.address import Address, Contact, FullContact
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment


def date_to_datetime(d: date) -> datetime:
    return datetime.combine(d, datetime.min.time())


def create_postage_details(shipment: Shipment, service_code):
    send_to = (
        SendNotifcationsTo.RECIPIENT if shipment.direction == ShipDirection.OUTBOUND else SendNotifcationsTo.BILLING
    )
    return PostageDetailsRequest(
        service_code=service_code,
        send_notifications_to=send_to,
        receive_email_notification=True,
        receive_sms_notification=True,
    )


def create_packages(*, num_parcels: int, package_format: PackageFormat):
    return [
        ShipmentPackageRequest(
            weight_in_grams=10000,
            package_format_identifier=package_format,
        )
        for _ in range(num_parcels)
    ]


def rm_address_from_agnostic_fc(full_contact: FullContact) -> AddressRM:
    return AddressRM(
        full_name=full_contact.contact.contact_name,
        company_name=full_contact.address.business_name,
        address_line1=full_contact.address.address_lines[0],
        address_line2=full_contact.address.address_lines[1] if len(full_contact.address.address_lines) > 1 else None,
        address_line3=full_contact.address.address_lines[2] if len(full_contact.address.address_lines) > 2 else None,
        city=full_contact.address.town,
        county=full_contact.address.county,
        postcode=full_contact.address.postcode,
        country_code=full_contact.address.country,
    )


def rm_recipient_details_from_agnostic_fc(full_contact: FullContact) -> ContactRM:
    return ContactRM(
        address=rm_address_from_agnostic_fc(full_contact),
        phone_number=full_contact.contact.mobile_phone or full_contact.contact.phone_number,
        email_address=full_contact.contact.email_address,
    )


def rm_billing_details_from_fc(full_contact: FullContact):
    return BillingDetailsRequest(
        address=rm_address_from_agnostic_fc(full_contact),
        phone_number=full_contact.contact.phone_number,
        email_address=full_contact.contact.email_address,
    )


def full_contact_from_rm(recipient: ContactRM) -> FullContact:
    return FullContact(
        contact=Contact(
            contact_name=recipient.address.full_name,
            phone_number=recipient.phone_number,
            email_address=recipient.email_address,
            mobile_phone=recipient.phone_number,
        ),
        address=Address(
            business_name=recipient.address.company_name,
            address_lines=[
                line
                for line in [
                    recipient.address.address_line1,
                    recipient.address.address_line2,
                    recipient.address.address_line3,
                ]
                if line
            ],
            town=recipient.address.city,
            county=recipient.address.county,
            postcode=recipient.address.postcode,
            country=recipient.address.country_code,
        ),
    )



from __future__ import annotations


from parcelforce_expresslink.models.address import AddressBase, AddressRecipient, AddressSender
from parcelforce_expresslink.models.contact import Contact as ContactPF, ContactSender
from parcelforce_expresslink.models.shipment import Shipment as ShipmentPF

from shipaw.models.address import Address as AddressAgnost, Contact as ContactAgnost, FullContact
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment as ShipmentAgnost


def address_from_agnostic[addr_type: AddressBase](
    address: AddressAgnost, cls: type[addr_type] = AddressRecipient
) -> addr_type:
    return cls(
        address_line1=address.address_lines[0],
        address_line2=address.address_lines[1] if len(address.address_lines) > 1 else None,
        address_line3=address.address_lines[2] if len(address.address_lines) > 2 else None,
        town=address.town,
        postcode=address.postcode,
        country=address.country,
    )


def address_from_agnostic_fc[addr_type: AddressBase](cls: type[addr_type], full_contact: FullContact) -> addr_type:
    return address_from_agnostic(full_contact.address, cls)


def contact_from_agnostic_fc[contact_type: ContactPF](
    cls: type[contact_type], full_contact: FullContact
) -> contact_type:
    return cls(
        business_name=full_contact.address.business_name,
        contact_name=full_contact.contact.contact_name,
        email_address=full_contact.contact.email_address,
        mobile_phone=full_contact.contact.mobile_phone,
    )


def parcelforce_full_contact(contact: ContactPF, address: AddressBase) -> FullContact:
    return FullContact(
        address=AddressAgnost(
            business_name=contact.business_name,
            address_lines=[
                line for line in [address.address_line1, address.address_line2, address.address_line3] if line
            ],
            town=address.town,
            postcode=address.postcode,
            country=address.country,
        ),
        contact=ContactAgnost(
            contact_name=contact.contact_name,
            email_address=contact.email_address,
            mobile_phone=contact.mobile_phone,
        ),
    )


def split_string_into_chunks(s, chunk_size):
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


def ref_dict_from_str(ref_string: str) -> dict[str, str]:
    refs = split_string_into_chunks(ref_string, 24)
    if len(refs) > 5:
        raise ValueError('Reference too long, maximum 120 characters allowed')
    ref_nums = {f'reference_number{i}': ref for i, ref in enumerate(refs, start=1)}
    return ref_nums


def parcelforce_shipment_to_agnostic(shipment: ShipmentPF) -> ShipmentAgnost:
    sender = None
    if shipment.direction == ShipDirection.INBOUND:
        address = shipment.collection_info.collection_address
        contact = shipment.collection_info.collection_contact
        sender = parcelforce_full_contact(contact, address)
    elif shipment.direction == ShipDirection.DROPOFF:
        address = shipment.sender_address
        contact = shipment.sender_contact
        sender = parcelforce_full_contact(contact, address)

    return ShipmentAgnost(
        recipient=parcelforce_full_contact(shipment.recipient_contact, shipment.recipient_address),
        sender=sender,
        boxes=shipment.total_number_of_parcels,
        shipping_date=shipment.shipping_date,
        direction=shipment.direction,
        reference=', '.join(
            filter(
                None,
                [
                    shipment.reference_number1,
                    shipment.reference_number2,
                    shipment.reference_number3,
                    shipment.reference_number4,
                    shipment.reference_number5,
                ],
            )
        ),
    )


def add_sender(ship_pf, shipment):
    ship_pf.sender_contact = contact_from_agnostic_fc(ContactSender, shipment.sender)
    ship_pf.sender_address = address_from_agnostic_fc(AddressSender, shipment.sender)


def convert_shipment_by_direction(ship_pf: ShipmentPF, shipment: ShipmentAgnost):
    if shipment.direction == ShipDirection.OUTBOUND:
        return
    add_sender(ship_pf, shipment)  # make dropoff
    if shipment.direction == ShipDirection.INBOUND:
        ship_pf.change_sender_to_collection()  # make inbound collection


def join_refs(refs: dict[str, str]) -> str:
    refs = [refs.get(f'reference_number{i + 1}', '') for i in range(len(refs))]
    return ''.join(refs).strip()



from pathlib import Path
from typing import cast
from urllib.error import HTTPError

from combadge.core.errors import BackendError
from fastapi import APIRouter, Body, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.staticfiles import StaticFiles

from shipaw.config import ShipawSettings
from shipaw.fapi.alerts import Alert, AlertType, Alerts, check_royal_mail, maybe_alert_phone_number
from shipaw.fapi.backend import convert_choice, maybe_alert_apc, try_book_shipment, try_get_write_label
from shipaw.fapi.form_data import provider_from_form, shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import AddressRequest, ShipmentRequest
from shipaw.fapi.responses import ShipawTemplate, ShipawTemplateResponse
from shipaw.models.address import Address, AddressChoice as AddressChoiceAgnost
from shipaw.models.logging import log_obj
from shipaw.models.shipment import Shipment
from shipaw.providers.parcelforce.parcelforce_funcs import (
    address_from_agnostic,
)
from shipaw.providers.parcelforce.parcelforce_provider import ParcelforceShippingProvider
from shipaw.providers.provider_abc import ProviderName
from shipaw.providers.registry import PROVIDER_REGISTER

router = APIRouter()
router.mount('/static', StaticFiles(directory=str(ShipawSettings.from_env().static_dir)), name='static')


def get_version():
    from importlib.metadata import version, PackageNotFoundError

    try:
        return version('shipaw')
    except PackageNotFoundError:
        return 'unknown'


async def notify_version():
    alerts = Alerts.empty()
    live = ShipawSettings.from_env().shipper_live
    if live:
        live_msg = 'Live Mode - Real Shipments will be booked'
        notification_type = AlertType.WARNING
    else:
        live_msg = 'Test Mode - No Shipments will be booked'
        notification_type = AlertType.NOTIFICATION
    if not live and ProviderName.ROYAL_MAIL in PROVIDER_REGISTER:
        live_msg += ' EXCEPT ROYAL MAIL!!'
        notification_type = AlertType.WARNING

    msg = f'Shipaw Version {get_version()} is in {live_msg}'
    logger.warning(msg)
    alerts += Alert(message=msg, type=notification_type)
    return alerts


def notify_dev() -> Alerts:
    alerts = Alerts.empty()
    if any(['prdev' in str(_).lower() for _ in Path(__file__).parents]):
        msg = '"prdev" in cwd tree - BETA MODE - This is a development version'
        logger.warning(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)
    return alerts


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def shipping_form_api(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj(shipment, 'Shipment received at /ship_form:')

    alerts: Alerts = request.app.alerts
    alerts += notify_dev()
    alerts += await notify_version()

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    # check phone number
    alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    # check not apc + dropoff
    alerts += await maybe_alert_apc(shipment_request)
    # check if royal mail and not standard service
    alerts += await check_royal_mail(shipment_request)

    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response = await try_book_shipment(shipment_request)
    await try_get_write_label(shipment_request, shipment_response)

    if shipment_response.alerts.errors:
        return await errored_shipment(shipment_response)

    log_obj(shipment_response, 'Shipment Booked')
    await try_get_write_label(shipment_request, shipment_response)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'shipment_request': shipment_request, 'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


async def errored_shipment(shipment_response):
    log_obj(shipment_response.alerts, 'Errors booking shipment:')
    alerts = shipment_response.alerts
    shipment_response.template = ShipawTemplate(
        template_path='/alerts.html',
        context={'alerts': alerts},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.post('/addr_choices', response_model=list[AddressChoiceAgnost], response_class=JSONResponse)
async def get_addr_choices_api(
    request: Request,
    body: AddressRequest = Body(...),
) -> list[AddressChoiceAgnost]:
    """Fetch candidate address choices for a postcode, optionally scored by closeness to provided address.
    Hardcoded to use Parcelforce provider for now - APC does not provide address lookup.

    Args:
        request: Request - FastAPI request object
        body: Address - request body containing postcode and optional address
    """

    parcelforce_ = PROVIDER_REGISTER.get(ProviderName.PARCELFORCE)
    if not parcelforce_:
        raise HTTPError('Parcelforce Provider not available - unable to fetch address candidates')
    p: ParcelforceShippingProvider = cast(ParcelforceShippingProvider, parcelforce_)
    client = p.client
    postcode = body.postcode
    address_agnost = body.address
    pf_address = address_from_agnostic(address_agnost) if address_agnost else None
    # log_obj(pf_address, 'Address received at /cand:')

    try:
        res_pf = client.get_choices(postcode=postcode, address=pf_address)
        res = [await convert_choice(_) for _ in res_pf]
        return res

    except BackendError as e:
        alert = Alert(
            message=f'Error fetching candidates: {e}',
            type=AlertType.ERROR,
        )
        request.app.alerts += alert
        logger.warning(f'Error fetching candidates: {e}')
        addr = Address(address_lines=['ERROR:', str(e)], town='Error', postcode='Error', business_name='Error')
        chc = AddressChoiceAgnost(address=addr, score=0)
        return [chc]


@router.get('/providers', response_class=JSONResponse)
async def get_providers():
    logger.warning('hit PROVIDERS')
    provider_response = {_.title(): _ for _ in PROVIDER_REGISTER.keys()}
    return JSONResponse(provider_response)


@router.get('/provider_services/{provider_name}', response_class=JSONResponse)
async def provider_services(provider_name: str):
    provider = await provider_from_form(provider_name)
    services = provider.services_as_dict()
    serv = {k.title(): v for k, v in services.items()}
    return JSONResponse(serv)


@router.get('/provider_directions/{provider_name}', response_class=JSONResponse)
async def provider_directions(provider_name: str):
    provider = await provider_from_form(provider_name)
    directions = provider.valid_directions
    dir_dict = {_.value: _.value for _ in directions}
    return JSONResponse(dir_dict)



import json
import pprint
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING
from collections.abc import Sequence

from loguru import logger
from pydantic import BaseModel

from shipaw.config import ShipawSettings

if TYPE_CHECKING:
    from shipaw.fapi.requests import ShipmentRequest
    from shipaw.fapi.responses import ShipmentResponse


def ndlog_dict(data: dict, ndjson_file: Path | None = None):
    ndjson_file = ndjson_file or ShipawSettings.from_env().ndjson_log_file
    with open(ndjson_file, 'a') as jf:
        print(json.dumps(data, separators=(',', ':')), file=jf)


def log_obj_text(obj: BaseModel, message: str = '', *, level: str = 'DEBUG', logger_=logger):
    message = message or obj.__class__.__name__
    logger_.log(
        level,
        message
        + ':\n'
        + pprint.pformat(
            obj.model_dump(
                mode='json',
                exclude={
                    'label_data': ...,
                    'response': {'label_data'},
                },
            ),
            indent=2,
        ),
    )


def log_obj_json(obj: BaseModel, message: str = '', *, ndjson_file=None):
    ndjson_file = ndjson_file or ShipawSettings.from_env().ndjson_log_file
    timestamp = datetime.now().isoformat(timespec='seconds')
    logdict = {
        'data_type': type(obj).__name__,
        'timestamp': timestamp,
        'message': message,
        'obj_data': obj.model_dump(mode='json', exclude={'label_data': ..., 'response': {'label_data'}}),
    }
    ndlog_dict(logdict, ndjson_file=ndjson_file)


def log_obj(
    obj: BaseModel,
    message: str = None,
    level: str = 'DEBUG',
    logger_=logger,
    ndjson_file=None,
):
    ndjson_file = ndjson_file or ShipawSettings.from_env().ndjson_log_file
    log_obj_text(obj, message, level=level, logger_=logger_)
    log_obj_json(obj, message, ndjson_file=ndjson_file)


def log_booked_shipment(request: 'ShipmentRequest', response: 'ShipmentResponse'):
    from shipaw.fapi.shipment_booking import ShipmentBooking

    conversation = ShipmentBooking(request=request, response=response)
    ndlog_dict(conversation.model_dump(mode='json', exclude={'response': {'label_data'}}))


def log_objs(objs: Sequence[BaseModel], message: str = None):
    if message:
        logger.debug(message + ':\n')
    for obj in objs:
        log_obj_text(obj)



from shipaw.providers.apc import APCShippingProvider, APCSettings
from shipaw.providers.parcelforce import ParcelforceShippingProvider, ParcelforceSettings
from shipaw.providers.royal_mail import RoyalMailProvider, RoyalMailSettings

__all__ = [
    'APCSettings',
    'APCShippingProvider',
    'ParcelforceSettings',
    'ParcelforceShippingProvider',
    'RoyalMailProvider',
    'RoyalMailSettings',
]



from .providers import APCSettings, APCShippingProvider, ParcelforceSettings, ParcelforceShippingProvider

__all__ = [
    'APCSettings',
    'APCShippingProvider',
    'ParcelforceSettings',
    'ParcelforceShippingProvider',
]



from apc_hypaship.error import apc_http_status_alerts
from httpx import HTTPStatusError
from loguru import logger
from parcelforce_expresslink.models.address import AddressChoice as AddressChoicePF
from parcelforce_expresslink.models.contact import Contact as ContactPF

from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.address import AddressChoice as AddressChoiceAgnost
from shipaw.models.ship_types import ShipDirection
from shipaw.providers.parcelforce.parcelforce_funcs import parcelforce_full_contact
from shipaw.providers.provider_abc import ProviderName


async def try_book_shipment(shipment_request: ShipmentRequest) -> ShipmentResponse:
    shipment_response = ShipmentResponse(shipment=shipment_request.shipment)
    try:
        shipment_response = shipment_request.provider.book_shipment_agnostic(shipment_request)

    except HTTPStatusError as e:
        await maybe_apc_response_error(e, shipment_request, shipment_response)

    except Exception as e:
        logger.exception(f'Error booking shipment: {e}')
        shipment_response.alerts += Alert.from_exception(e)

    return shipment_response


async def try_get_write_label(request: ShipmentRequest, response: ShipmentResponse):
    if not response.label_data:
        await try_get_label_data(request, response)
    await try_write_label(response)


async def try_get_label_data(request: ShipmentRequest, response: ShipmentResponse) -> None:
    try:
        if response.label_data is not None:
            logger.info('Label data already present, not fetching')
        else:
            if response.shipment_num:
                response.label_data = await request.provider.wait_fetch_label_as(response.shipment_num)
            else:
                logger.warning('No shipment number to fetch label data')

    except HTTPStatusError as e:
        await maybe_apc_response_error(e, request, response)

    except Exception as e:
        logger.exception('Error getting label data')
        response.alerts += Alert.from_exception(e)


async def try_write_label(response: ShipmentResponse):
    try:
        await response.write_label_file()
    except Exception as e:
        logger.exception(f'Error writing label file: {e}')
        response.alerts += Alert.from_exception(e)


async def maybe_apc_response_error(e: HTTPStatusError, shipment_request, shipment_response):
    if shipment_request.provider_name == ProviderName.APC:
        for alert in await apc_http_status_alerts(e):
            shipment_response.alerts += alert
    else:
        logger.exception(e)
        shipment_response.alerts += Alert.from_exception(e)


async def maybe_alert_apc(shipment_request):
    alerts = Alerts.empty()
    if (
        shipment_request.provider_name == ProviderName.APC
        and shipment_request.shipment.direction == ShipDirection.DROPOFF
    ):
        alerts += Alert(
            message='APC does not support drop-off shipments - please select Outbound or Inbound Collection',
            type=AlertType.ERROR,
        )
    return alerts


async def convert_choice(choice: AddressChoicePF) -> AddressChoiceAgnost:
    fc = parcelforce_full_contact(contact=ContactPF.empty(), address=choice.address)
    return AddressChoiceAgnost(address=fc.address, score=choice.score)



from parcelforce_expresslink.config import ParcelforceSettings
from parcelforce_expresslink.expresslink_client import ParcelforceClient

from shipaw.models.logging import log_obj


def cancel_parcelforce(ship_number):
    live_settings = ParcelforceSettings.from_env('PARCELFORCE_ENV')
    client = ParcelforceClient(settings=live_settings)
    resp = client.cancel_shipment(ship_number)
    log_obj(resp)
    return resp


if __name__ == '__main__':
    # res = cancel_parcelforce()

    ...


import contextlib
import os

from fastapi import FastAPI, Query, responses
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from loguru import logger
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.staticfiles import StaticFiles

from shipaw.config import ShipawSettings
from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.routes_api import router as json_router
from shipaw.fapi.routes_html import router as html_router


@contextlib.asynccontextmanager
async def lifespan(app_: FastAPI):
    try:
        # app_.settings = ShipawSettings.from_env()
        # set_pf_env()
        # pythoncom.CoInitialize()
        # with sqm.Session(am_db.ENGINE) as session:
        #     pf_shipper = ELClient()
        #     populate_db_from_cmc(session, pf_shipper)
        yield

    finally:
        # pythoncom.CoUninitialize()

        ...


app = FastAPI(lifespan=lifespan)
app.mount('/static', StaticFiles(directory=str(ShipawSettings.from_env().static_dir)), name='static')
app.include_router(json_router, prefix='/api')
app.include_router(html_router)
app.ship_live = ShipawSettings.from_env().shipper_live
app.alerts = Alerts.empty()


@app.exception_handler(RequestValidationError)
async def request_validation_exception_handler(request: Request, exc: RequestValidationError):
    errors = exc.errors()
    msg2 = ''
    for err in errors:
        if type := err.get('type'):
            msg2 += type + ' '
        if loc := err.get('loc'):
            msg2 += f'in {loc} '
        if ctx := err.get('ctx'):
            if reason := ctx.get('reason'):
                msg2 += f': {reason} '
        if input_ := err.get('input'):
            msg2 += f'. Input = {input_} '

    logger.error(msg2)
    alerts = Alerts(alert=[Alert(code=1, message=msg2, type=AlertType.ERROR)])
    return JSONResponse(
        status_code=422,
        content={'detail': jsonable_encoder(exc.errors()), 'alerts': alerts.model_dump(mode='json'), 'message': msg2},
    )


@app.get('/robots.txt', response_class=responses.PlainTextResponse)
async def robots_txt() -> str:
    return 'User-agent: *\nAllow: /'


@app.get('/favicon.ico', include_in_schema=False)
async def favicon_ico():
    return responses.RedirectResponse(url='/static/favicon.svg')


@app.get('/', response_class=JSONResponse)
async def base():
    return JSONResponse(content={'status': 'ok'})


@app.get('/open-file', response_class=HTMLResponse)
async def open_file(filepath: str = Query(...)):
    os.startfile(filepath)
    return HTMLResponse(content='<span>Re</span>')



from typing import ClassVar, override

from loguru import logger
from royal_mail_click_and_drop import (
    CreateOrdersRequest,
    CreateOrdersResponse,
)
from royal_mail_click_and_drop.config import RoyalMailSettings
from royal_mail_click_and_drop.models.create_orders_request import CreateOrderRequest
from royal_mail_click_and_drop.models.shipment_package_request import PackageFormat
from royal_mail_click_and_drop.v2.client import RoyalMailClient
from royal_mail_click_and_drop.v2.services import RoyalMailServiceCode

from shipaw.config import ShipawSettings
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.logging import log_obj
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ProviderName, ShippingProvider
from shipaw.providers.registry import register_provider_type
from shipaw.providers.royal_mail.royal_mail_funcs import (
    create_packages,
    create_postage_details,
    date_to_datetime,
    full_contact_from_rm,
    rm_billing_details_from_fc,
    rm_recipient_details_from_agnostic_fc,
)


@register_provider_type
class RoyalMailProvider(ShippingProvider):
    settings: RoyalMailSettings

    name: ClassVar[ProviderName] = ProviderName.ROYAL_MAIL
    settings_type: ClassVar[type[RoyalMailSettings]] = RoyalMailSettings
    service_codes_type: ClassVar[type[RoyalMailServiceCode]] = RoyalMailServiceCode
    default_service: ClassVar[RoyalMailServiceCode] = RoyalMailServiceCode.EXPRESS_24
    valid_directions: ClassVar[list[ShipDirection]] = [ShipDirection.OUTBOUND]

    _client: RoyalMailClient | None = None

    @override
    def is_sandbox(self) -> bool:
        # Royal mail do not have a test environment. that's fun
        return False

    @property
    @override
    def client(self) -> RoyalMailClient:
        if self._client is None:
            if self.settings is None:
                raise ValueError('Settings must be set before using the client')
            self._client = RoyalMailClient(settings=self.settings)
        return self._client

    @classmethod
    @override
    def agnostic_shipment(cls, orders_req: CreateOrdersRequest) -> Shipment:
        order = orders_req.items[0]
        return Shipment(
            recipient=full_contact_from_rm(order.recipient),
            boxes=len(order.packages),
            shipping_date=order.planned_despatch_date.date(),
            direction=ShipDirection.OUTBOUND,
            reference=order.order_reference,
        )

    @override
    def provider_shipment(self, shipment: Shipment, service_code: RoyalMailServiceCode) -> CreateOrdersRequest:
        if shipment.direction != ShipDirection.OUTBOUND:
            # todo: implement inbound / dropoff
            raise NotImplementedError('only outbound shipments are supported for Royal Mail')
        shipaw_settings = ShipawSettings.from_env('SHIPAW_ENV')
        billing_details = rm_billing_details_from_fc(shipaw_settings.full_contact)
        postage_details = create_postage_details(shipment=shipment, service_code=service_code)

        return CreateOrdersRequest(
            items=[
                CreateOrderRequest(
                    order_reference=shipment.reference,
                    postage_details=postage_details,
                    billing=billing_details,
                    recipient=rm_recipient_details_from_agnostic_fc(shipment.recipient),
                    order_date=date_to_datetime(shipment.shipping_date),
                    planned_despatch_date=date_to_datetime(shipment.shipping_date),
                    subtotal=0,
                    shipping_cost_charged=0,
                    total=0,
                    packages=create_packages(num_parcels=shipment.boxes, package_format=PackageFormat.PARCEL),
                )
            ]
        )

    def provider_shipment_request(self, shipment_request: ShipmentRequest) -> CreateOrdersRequest:
        provider_service = self.service_codes_type(shipment_request.service_code)
        shipment_rm = self.provider_shipment(shipment_request.shipment, provider_service)
        log_obj(shipment_rm, 'Royal Mail Shipment Request')
        return shipment_rm

    @override
    def book_shipment_agnostic(self, shipment_request: ShipmentRequest) -> ShipmentResponse:
        shipment = shipment_request.shipment
        provider_shipment_request = self.provider_shipment_request(shipment_request)
        resp = self.client.book_shipment(provider_shipment_request)
        logger.warning(f'BOOKED SHIPMENTS: {resp.created_orders_idents}')
        return self.build_booking_response(resp, shipment)

    def build_booking_response(self, rm_response: CreateOrdersResponse, shipment):
        """without label data"""
        if rm_response.errors_count > 0:
            raise RuntimeError()
        order = rm_response.created_orders[0]
        track_num = order.tracking_number if order.tracking_number else ''
        success = rm_response.errors_count == 0
        return ShipmentResponse(
            shipment=shipment,
            shipment_num=str(order.order_identifier),
            tracking_link=self.settings.tracking_link(track_num),
            data=rm_response.model_dump(),
            status='Success' if success else 'FAIL',
            success=success,
        )

    @override
    def fetch_label_content(self, shipment_num: str) -> bytes:
        return self.client.get_label_content(shipment_num)
